/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: JORAMHelper.java 2344 2009-06-02 19:23:34Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server.impl.jms;

import java.net.ConnectException;
import java.util.Arrays;
import java.util.List;

import javax.jms.Topic;

import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.joram.client.jms.admin.AdminModule;
import org.ow2.cmi.controller.server.ClusterView;
import org.ow2.cmi.controller.server.Domain;
import org.ow2.cmi.controller.server.ServerView;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Helper to create topics on a JORAM server.
 * @author Loris Bouzonnet
 */
public final class JORAMHelper {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(JORAMHelper.class);


    /**
     * Helper class: no public constructor.
     */
    private JORAMHelper() {
        // Do nothing
    }

    public static Topic createTopic(
            final String topicName, final Topic ourTopic, final Topic superTopic) throws JORAMServerException {
        org.objectweb.joram.client.jms.Topic topic;
        try {
            topic = org.objectweb.joram.client.jms.Topic.create(topicName);
            topic.setFreeReading();
            topic.setFreeWriting();
            if(superTopic != null && superTopic instanceof org.objectweb.joram.client.jms.Topic) {
                topic.setParent(
                        (org.objectweb.joram.client.jms.Topic) superTopic);
            } else if(ourTopic != null && ourTopic instanceof org.objectweb.joram.client.jms.Topic) {
                topic.addClusteredTopic(
                        (org.objectweb.joram.client.jms.Topic) ourTopic);
            }
        } catch (ConnectException e) {
            logger.error("Unable to create the topic with name " + topicName, e);
            throw new JORAMServerException("Unable to create the topic with name " + topicName, e);
        } catch (AdminException e) {
            logger.error("Unable to create the topic with name " + topicName, e);
            throw new JORAMServerException("Unable to create the topic with name " + topicName, e);
        }
        return topic;
    }

    public static void checkMyTopic(final Topic myTopic, final Topic ourTopic,
            final Topic superTopic) throws JORAMServerException {
        if(superTopic != null && superTopic instanceof org.objectweb.joram.client.jms.Topic) {
            try {
                Topic topic =
                    ((org.objectweb.joram.client.jms.Topic) myTopic).getHierarchicalFather();
                if(!superTopic.equals(topic)) {
                    logger.error("Not a hierarchical topic as expected.");
                    throw new JORAMServerException("Not a hierarchical topic as expected.");
                }
            } catch (ConnectException e) {
                logger.error("Unable to check the use of hierarchical topic", e);
                throw new JORAMServerException("Unable to check the use of hierarchical topic", e);
            } catch (AdminException e) {
                logger.error("Unable to check the use of hierarchical topic", e);
                throw new JORAMServerException("Unable to check the use of hierarchical topic", e);
            }
        } else if(ourTopic != null && ourTopic instanceof org.objectweb.joram.client.jms.Topic) {
            try {
                List<?> fellows =
                    ((org.objectweb.joram.client.jms.Topic) myTopic).getClusterFellows();
                if(!fellows.contains(ourTopic)) {
                    logger.error("Not a clustered topic as expected.");
                    throw new JORAMServerException("Not a clustered topic as expected.");
                }
            } catch (ConnectException e) {
                logger.error("Unable to check the use of clustered topic", e);
                throw new JORAMServerException("Unable to check the use of clustered topic", e);
            } catch (AdminException e) {
                logger.error("Unable to check the use of clustered topic", e);
                throw new JORAMServerException("Unable to check the use of clustered topic", e);
            }
        }
    }

    /**
     * @param config
     * @throws JORAMServerException
     */
    public static void setDistributedConfig(final Config config, final ClusterView clusterView) throws JORAMServerException {
        String domainName = config.getDomainName();
        if(domainName == null) {
            logger.error("Unable to set a distributed config without domain name!");
            throw new JORAMServerException("Unable to set a distributed config without domain name!");
        }
        try {
            if(!Arrays.asList(AdminModule.getDomainNames(config.getServerId())).contains(domainName)) {

                if (config.getDomainPort() == null) {
                    logger.error("Unable to add a domain without its domain name");
                    throw new JORAMServerException(
                    "Unable to add a domain without its domain name");
                }
                AdminModule.addDomain(domainName,
                        config.getServerId(), config.getDomainPort());
                if (clusterView != null) {
                    for (ServerView serverView : clusterView.getServerViews(domainName)) {
                        Domain domain = serverView.getDomain(domainName);
                        AdminModule.addServer(
                                serverView.getId(), serverView.getServerRef().getHostname(),
                                domainName, domain.getPort(), serverView.getServerName());
                    }
                }
            }
        } catch (Exception e) {
            logger.error("Unable to administrate the JORAM server", e);
            throw new JORAMServerException("Unable to administrate the JORAM server", e);
        }
    }

    /**
     * @return
     * @throws JORAMServerException
     */
    public static String getServerName() throws JORAMServerException {
        try {
            return AdminModule.getLocalServer().getName();
        } catch (ConnectException e) {
            logger.error("Unable to retrieve the local server name", e);
            throw new JORAMServerException("Unable to retrieve the local server name", e);
        } catch (AdminException e) {
            logger.error("Unable to retrieve the local server name", e);
            throw new JORAMServerException("Unable to retrieve the local server name", e);
        }
    }

}
